==> go-testing-202.8194.19-sources.jar: com/goide/testing/GoTestJsonEventsProcessor.java <==
// Copyright 2000-2020 JetBrains s.r.o.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.goide.testing;

import com.goide.GoStringUtil;
import com.google.gson.Gson;
import com.google.gson.JsonSyntaxException;
import jetbrains.buildServer.messages.serviceMessages.ServiceMessageTypes;
import jetbrains.buildServer.messages.serviceMessages.ServiceMessageVisitor;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

public class GoTestJsonEventsProcessor {
  private static final String NOT_EQUAL_PATTERN = "Error:\\s+Not equal:.*\n";
  private static final String WINDOWS_PTY_PREFIX = "\u001b[0m";

  @NotNull
  private final GoTestEventProcessor myProcessor;

  private final Set<String> myEmptySuites = new HashSet<String>();
  private final Set<String> myStartedPackages = new HashSet<String>();
  private final Set<String> myStartedTests = new HashSet<String>();
  private final Set<String> myTestsWithIgnoredChildren = new HashSet<String>();
  private final Map<String, ComparisonFailures> myComparisonFailures = new HashMap<String, ComparisonFailures>();
  private final String myDefaultImportPath;
  private final Object myFailTestSync = new Object();
  private volatile String myCurrentFailedImportPath;
  private final StringBuilder myPendingInvalidJson = new StringBuilder();
  private final boolean myHandleSplitLines;
  private final boolean myRestoreTrimmedSpaces;

  public GoTestJsonEventsProcessor(@NotNull GoTestEventProcessor processor) {
    this("", processor, false, false);
  }

  public GoTestJsonEventsProcessor(@NotNull String defaultImportPath,
                                   @NotNull GoTestEventProcessor processor,
                                   boolean handleSplitLines,
                                   boolean restoreTrimmedSpaces) {
    myProcessor = processor;
    myDefaultImportPath = defaultImportPath;
    myHandleSplitLines = handleSplitLines;
    myRestoreTrimmedSpaces = restoreTrimmedSpaces;
  }

  public void clearState() {
    myEmptySuites.clear();
    myStartedPackages.clear();
    myStartedTests.clear();
    myTestsWithIgnoredChildren.clear();
    myComparisonFailures.clear();
    myCurrentFailedImportPath = null;
    myPendingInvalidJson.setLength(0);
  }

  public boolean processLine(@NotNull String text, @NotNull OutputType outputType, ServiceMessageVisitor visitor) {
    Gson gson = new Gson();
    String textWithoutWindowsPrefix = text;
    GoTestEvent event;
    if (myPendingInvalidJson.length() > 0) {
      appendPendingInvalidJson(text);
      try {
        event = gson.fromJson(myPendingInvalidJson.toString(), GoTestEvent.class);
      }
      catch (JsonSyntaxException e) {
        // continue collecting output
        return true;
      }
      myPendingInvalidJson.setLength(0);
    }
    else {
      if (text.startsWith(WINDOWS_PTY_PREFIX + "{\"Time\":")) {
        textWithoutWindowsPrefix = text.substring(WINDOWS_PTY_PREFIX.length());
      }
      try {
        event = gson.fromJson(textWithoutWindowsPrefix, GoTestEvent.class);
      }
      catch (JsonSyntaxException e) {
        if (myHandleSplitLines && looksLikeJson(textWithoutWindowsPrefix)) {
          appendPendingInvalidJson(textWithoutWindowsPrefix);
          return true;
        }
        event = null;
      }
    }
    if (event == null) {
      return handleNonJsonOutput(text, outputType, visitor);
    }
    finishNonJsonOutputIfNeeded(visitor);

    Action action = event.getAction();
    NameAndParentTestName nameAndParentTestName = getNames(event.testName, event.importPath);
    String name = nameAndParentTestName.testName;
    if (action == Action.OUTPUT) {
      if (name != null) {
        if (event.isPackage() && myStartedPackages.add(event.importPath)) {
          if (!startPackage(event, visitor)) {
            return false;
          }
        }
        processComparisonFailure(event.getNodeId(), event.output);
        return output(name, event.getNodeId(), event.output, outputType, visitor);
      }
      else {
        ServiceMessageBuilder message = new ServiceMessageBuilder(ServiceMessageTypes.MESSAGE)
          .attribute("text", GoStringUtil.notNullize(event.output))
          .attribute("status", outputType == OutputType.STDERR ? "ERROR" : "NORMAL");
        return myProcessor.processEvent(message.build(), outputType, visitor);
      }
    }

    if (name == null || action == Action.CONT || action == Action.PAUSE) {
      return true;
    }

    if (action == Action.RUN && event.isTest()) {
      if (event.importPath != null && myStartedPackages.add(event.importPath)) {
        if (!startPackage(event, visitor)) {
          return false;
        }
      }
      return startTest(event, visitor);
    }

    if (action == Action.PASS || action == Action.FAIL || action == Action.SKIP) {
      if (event.isPackage() && myStartedPackages.add(event.importPath)) {
        if (!startPackage(event, visitor)) {
          return false;
        }
      }

      boolean suite = event.isPackage() && myStartedPackages.remove(event.importPath);
      if (action == Action.FAIL && (!suite || myEmptySuites.contains(event.getNodeId()))) {
        ServiceMessageBuilder failTest = ServiceMessageBuilder.testFailed(name)
          .attribute("nodeId", event.getNodeId())
          .attribute("message", GoStringUtil.notNullize(event.output));
        ComparisonFailures comparisonFailure = myComparisonFailures.remove(event.getNodeId());
        if (comparisonFailure != null) {
          failTest.attribute("type", "comparisonFailure");
          failTest.attribute("expected", decodeComparisonValue(comparisonFailure.expected));
          failTest.attribute("actual", decodeComparisonValue(comparisonFailure.actual));
        }
        if (!myProcessor.processEvent(failTest.build(), OutputType.SYSTEM, visitor)) {
          return false;
        }
      }
      if (!suite && (action == Action.SKIP || (action == Action.PASS && myTestsWithIgnoredChildren.remove(event.getNodeId())))) {
        ServiceMessageBuilder skipTest = ServiceMessageBuilder.testIgnored(name)
          .attribute("nodeId", event.getNodeId())
          .attribute("message", GoStringUtil.notNullize(event.output));
        String parentNodeId = event.getParentNodeId(nameAndParentTestName.parentTestName);
        if (parentNodeId != null) {
          myTestsWithIgnoredChildren.add(parentNodeId);
        }
        if (!myProcessor.processEvent(skipTest.build(), OutputType.SYSTEM, visitor)) {
          return false;
        }
      }
      return finishTest(event, name, suite, visitor);
    }
    return false;
  }

  private void appendPendingInvalidJson(@NotNull String text) {
    String trimmed;
    if (text.endsWith("\r\n")) {
      trimmed = text.substring(0, text.length() - 2);
    }
    else if (text.length() > 0 && text.charAt(text.length() - 1) == '\n'){
      trimmed = text.substring(0, text.length() - 1);
    }
    else {
      trimmed = text;
    }
    if (myRestoreTrimmedSpaces) {
      // output may lose some whitespaces, see https://youtrack.jetbrains.com/issue/IDEA-236807
      int trimmedSpaces = 80 - trimmed.length();
      if (trimmedSpaces > 0) {
        trimmed += GoStringUtil.repeat(" ", trimmedSpaces);
      }
    }
    myPendingInvalidJson.append(trimmed);
  }

  private static boolean looksLikeJson(@NotNull String text) {
    return text.startsWith("{\"Time\":");
  }

  @NotNull
  private static String decodeComparisonValue(@Nullable String value) {
    String literalText = GoStringUtil.unescapeStringLiteralText(value);
    return literalText != null ? literalText : GoStringUtil.notNullize(value);
  }

  private void processComparisonFailure(@Nullable String nodeId, @Nullable String output) {
    if (GoStringUtil.isNotEmpty(nodeId) && GoStringUtil.isNotEmpty(output)) {
      String trimmedOutput = GoStringUtil.trimLeading(output);
      ComparisonFailures comparisonFailures = myComparisonFailures.get(nodeId);
      if (comparisonFailures != null) {
        if (trimmedOutput.startsWith("expected:")) {
          comparisonFailures.expected = GoStringUtil.trimStart(trimmedOutput, "expected:").trim();
        }
        else if (trimmedOutput.startsWith("actual  :")) {
          comparisonFailures.actual = GoStringUtil.trimStart(trimmedOutput, "actual  :").trim();
        }
      }
      else if (Pattern.compile(NOT_EQUAL_PATTERN).matcher(trimmedOutput).matches()) {
        myComparisonFailures.put(nodeId, new ComparisonFailures());
      }
    }
  }

  private boolean startPackage(@NotNull GoTestEvent event, ServiceMessageVisitor visitor) {
    assert event.importPath != null;
    myEmptySuites.add(event.getNodeId());
    return startTest(event, event.importPath, event.importPath, null, event.getPackageLocationString(myDefaultImportPath), true, visitor);
  }

  private boolean startTest(@NotNull GoTestEvent event, ServiceMessageVisitor visitor) {
    NameAndParentTestName names = getNames(event.testName, event.importPath);
    assert names.testName != null;
    String parentNodeId = event.getParentNodeId(names.parentTestName);
    return startTest(event, names.testName, event.getNodeId(), parentNodeId, event.getTestLocationString(myDefaultImportPath), false, visitor);
  }

  private boolean output(@NotNull String testName,
                         @Nullable String nodeId,
                         @Nullable String output,
                         @NotNull OutputType outputType,
                         ServiceMessageVisitor visitor) {
    ServiceMessageBuilder message = outputType == OutputType.STDERR ? ServiceMessageBuilder.testStdErr(testName)
                                                                    : ServiceMessageBuilder.testStdOut(testName);
    message.attribute("nodeId", nodeId).attribute("out", output);
    return myProcessor.processEvent(message.build(), outputType, visitor);
  }

  private boolean startTest(@NotNull GoTestEvent event,
                            @NotNull String testName,
                            @Nullable String nodeId,
                            @Nullable String parentNodeId,
                            @NotNull String locationString,
                            boolean suite,
                            ServiceMessageVisitor visitor) {
    myStartedTests.add(nodeId);
    myEmptySuites.remove(parentNodeId);
    ServiceMessageBuilder startTest = suite ? ServiceMessageBuilder.testSuiteStarted(testName)
                                            : ServiceMessageBuilder.testStarted(testName);
    startTest
      .attribute("nodeId", nodeId)
      .attribute("parentNodeId", GoStringUtil.notNullize(parentNodeId, "0"))
      .attribute("locationHint", locationString)
      .attribute("message", GoStringUtil.notNullize(event.output));
    return myProcessor.processEvent(startTest.build(), OutputType.SYSTEM, visitor);
  }

  private void finishNonJsonOutputIfNeeded(ServiceMessageVisitor visitor) {
    synchronized (myFailTestSync) {
      if (myCurrentFailedImportPath != null) {
        String nodeId = myCurrentFailedImportPath;
        myCurrentFailedImportPath = null;
        ServiceMessageBuilder failTest = ServiceMessageBuilder.testFailed(nodeId)
          .attribute("nodeId", nodeId)
          .attribute("message", "");
        myProcessor.processEvent(failTest.build(), OutputType.SYSTEM, visitor);
        ServiceMessageBuilder finishTest = ServiceMessageBuilder.testSuiteFinished(nodeId).attribute("nodeId", nodeId);
        myProcessor.processEvent(finishTest.build(), OutputType.SYSTEM, visitor);
      }
    }
  }

  private boolean finishTest(@NotNull GoTestEvent event, @NotNull String testName, boolean suite, ServiceMessageVisitor visitor) {
    ServiceMessageBuilder finishTest = suite ? ServiceMessageBuilder.testSuiteFinished(GoStringUtil.notNullize(event.importPath))
                                             : ServiceMessageBuilder.testFinished(testName);
    finishTest.attribute("nodeId", event.getNodeId()).attribute("duration", event.getElapsedTime());
    return myProcessor.processEvent(finishTest.build(), OutputType.SYSTEM, visitor);
  }

  private boolean handleNonJsonOutput(@NotNull String text, @NotNull OutputType type, ServiceMessageVisitor visitor) {
    synchronized (myFailTestSync) {
      if (text.startsWith("# ")) {
        myCurrentFailedImportPath = GoStringUtil.trimStart(text, "# ");
        ServiceMessageBuilder messageBuilder = ServiceMessageBuilder.testSuiteStarted(myCurrentFailedImportPath)
          .attribute("nodeId", myCurrentFailedImportPath)
          .attribute("parentNodeId", "0");
        if (!myProcessor.processEvent(messageBuilder.build(), type, visitor)) {
          return false;
        }
      }
      if (myCurrentFailedImportPath != null) {
        return output(myCurrentFailedImportPath, myCurrentFailedImportPath, text, type, visitor);
      }
      return false;
    }
  }

  enum Action {RUN, PASS, SKIP, FAIL, OUTPUT, CONT, PAUSE}

  private NameAndParentTestName getNames(@Nullable String testName, @Nullable String importPath) {
    String parentTestName = "";
    if (testName != null) {
      String notNullImportPath = GoStringUtil.notNullize(importPath);
      int lastSlash = testName.lastIndexOf('/');
      while (lastSlash > -1) {
        String beforeSlash = testName.substring(0, lastSlash);
        if (lastSlash < testName.length() - 1 && myStartedTests.contains(notNullImportPath + "#" + beforeSlash)) {
          testName = testName.substring(lastSlash + 1);
          parentTestName = beforeSlash;
          break;
        }
        lastSlash = testName.lastIndexOf('/', lastSlash - 1);
      }
    }
    return new NameAndParentTestName(testName != null ? testName : importPath, parentTestName);
  }

  private static class ComparisonFailures {
    public String expected;
    public String actual;
  }

  private static class NameAndParentTestName {
    public final String testName;
    public final String parentTestName;

    private NameAndParentTestName(@Nullable String testName, @Nullable String parentTestName) {
      this.testName = testName;
      this.parentTestName = parentTestName;
    }
  }
}

==> go-testing-203.8084.22-sources.jar: com/goide/testing/GoTestJsonEventsProcessor.java <==
// Copyright 2000-2020 JetBrains s.r.o.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.goide.testing;

import com.goide.GoStringUtil;
import com.google.gson.Gson;
import com.google.gson.JsonSyntaxException;
import jetbrains.buildServer.messages.serviceMessages.ServiceMessageTypes;
import jetbrains.buildServer.messages.serviceMessages.ServiceMessageVisitor;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

public class GoTestJsonEventsProcessor {
  private static final String NOT_EQUAL_PATTERN = "Error:\\s+Not equal:.*\n";
  private static final String WINDOWS_PTY_PREFIX = "\u001b[0m";

  @NotNull
  private final GoTestEventProcessor myProcessor;

  private final Set<String> myEmptySuites = new HashSet<String>();
  private final Set<String> myStartedPackages = new HashSet<String>();
  private final Set<String> myStartedTests = new HashSet<String>();
  private final Set<String> myTestsWithIgnoredChildren = new HashSet<String>();
  private final Map<String, ComparisonFailures> myComparisonFailures = new HashMap<String, ComparisonFailures>();
  private final String myDefaultImportPath;
  private final Object myFailTestSync = new Object();
  private volatile String myCurrentFailedImportPath;
  private final StringBuilder myPendingInvalidJson = new StringBuilder();
  private final boolean myHandleSplitLines;
  private final boolean myRestoreTrimmedSpaces;
  private Double myTotalTimeSeconds;

  public GoTestJsonEventsProcessor(@NotNull GoTestEventProcessor processor) {
    this("", processor, false, false);
  }

  public GoTestJsonEventsProcessor(@NotNull String defaultImportPath,
                                   @NotNull GoTestEventProcessor processor,
                                   boolean handleSplitLines,
                                   boolean restoreTrimmedSpaces) {
    myProcessor = processor;
    myDefaultImportPath = defaultImportPath;
    myHandleSplitLines = handleSplitLines;
    myRestoreTrimmedSpaces = restoreTrimmedSpaces;
  }

  public void clearState() {
    myEmptySuites.clear();
    myStartedPackages.clear();
    myStartedTests.clear();
    myTestsWithIgnoredChildren.clear();
    myComparisonFailures.clear();
    myCurrentFailedImportPath = null;
    myPendingInvalidJson.setLength(0);
  }

  public boolean processLine(@NotNull String text, @NotNull OutputType outputType, ServiceMessageVisitor visitor) {
    Gson gson = new Gson();
    String textWithoutWindowsPrefix = text;
    GoTestEvent event;
    if (myPendingInvalidJson.length() > 0) {
      appendPendingInvalidJson(text);
      try {
        event = gson.fromJson(myPendingInvalidJson.toString(), GoTestEvent.class);
      }
      catch (JsonSyntaxException e) {
        // continue collecting output
        return true;
      }
      myPendingInvalidJson.setLength(0);
    }
    else {
      if (text.startsWith(WINDOWS_PTY_PREFIX + "{\"Time\":")) {
        textWithoutWindowsPrefix = text.substring(WINDOWS_PTY_PREFIX.length());
      }
      try {
        event = gson.fromJson(textWithoutWindowsPrefix, GoTestEvent.class);
      }
      catch (JsonSyntaxException e) {
        if (myHandleSplitLines && looksLikeJson(textWithoutWindowsPrefix)) {
          appendPendingInvalidJson(textWithoutWindowsPrefix);
          return true;
        }
        event = null;
      }
    }
    if (event == null) {
      return handleNonJsonOutput(text, outputType, visitor);
    }
    finishNonJsonOutputIfNeeded(visitor);

    Action action = event.getAction();
    NameAndParentTestName nameAndParentTestName = getNames(event.testName, event.importPath);
    String name = nameAndParentTestName.testName;
    if (action == Action.OUTPUT) {
      if (name != null) {
        if (event.isPackage() && myStartedPackages.add(event.importPath)) {
          if (!startPackage(event, visitor)) {
            return false;
          }
        }
        processComparisonFailure(event.getNodeId(), event.output);
        return output(name, event.getNodeId(), event.output, outputType, visitor);
      }
      else {
        ServiceMessageBuilder message = new ServiceMessageBuilder(ServiceMessageTypes.MESSAGE)
          .attribute("text", GoStringUtil.notNullize(event.output))
          .attribute("status", outputType == OutputType.STDERR ? "ERROR" : "NORMAL");
        return myProcessor.processEvent(message.build(), outputType, visitor);
      }
    }

    if (name == null) {
      if (action == Action.PASS || action == Action.FAIL) {
        myTotalTimeSeconds = event.elapsed;
      }
      return true;
    }
    if (action == Action.CONT || action == Action.PAUSE) {
      return true;
    }

    if (action == Action.RUN && event.isTest()) {
      if (event.importPath != null && myStartedPackages.add(event.importPath)) {
        if (!startPackage(event, visitor)) {
          return false;
        }
      }
      return startTest(event, visitor);
    }

    if (action == Action.PASS || action == Action.FAIL || action == Action.SKIP) {
      if (event.isPackage() && myStartedPackages.add(event.importPath)) {
        if (!startPackage(event, visitor)) {
          return false;
        }
      }

      boolean suite = event.isPackage() && myStartedPackages.remove(event.importPath);
      if (action == Action.FAIL && (!suite || myEmptySuites.contains(event.getNodeId()))) {
        ServiceMessageBuilder failTest = ServiceMessageBuilder.testFailed(name)
          .attribute("nodeId", event.getNodeId())
          .attribute("message", GoStringUtil.notNullize(event.output));
        ComparisonFailures comparisonFailure = myComparisonFailures.remove(event.getNodeId());
        if (comparisonFailure != null) {
          failTest.attribute("type", "comparisonFailure");
          failTest.attribute("expected", decodeComparisonValue(comparisonFailure.expected));
          failTest.attribute("actual", decodeComparisonValue(comparisonFailure.actual));
        }
        if (!myProcessor.processEvent(failTest.build(), OutputType.SYSTEM, visitor)) {
          return false;
        }
      }
      if (!suite && (action == Action.SKIP || (action == Action.PASS && myTestsWithIgnoredChildren.remove(event.getNodeId())))) {
        ServiceMessageBuilder skipTest = ServiceMessageBuilder.testIgnored(name)
          .attribute("nodeId", event.getNodeId())
          .attribute("message", GoStringUtil.notNullize(event.output));
        String parentNodeId = event.getParentNodeId(nameAndParentTestName.parentTestName);
        if (parentNodeId != null) {
          myTestsWithIgnoredChildren.add(parentNodeId);
        }
        if (!myProcessor.processEvent(skipTest.build(), OutputType.SYSTEM, visitor)) {
          return false;
        }
      }
      return finishTest(event, name, suite, visitor);
    }
    return false;
  }

  private void appendPendingInvalidJson(@NotNull String text) {
    String trimmed;
    if (text.endsWith("\r\n")) {
      trimmed = text.substring(0, text.length() - 2);
    }
    else if (text.length() > 0 && text.charAt(text.length() - 1) == '\n'){
      trimmed = text.substring(0, text.length() - 1);
    }
    else {
      trimmed = text;
    }
    if (myRestoreTrimmedSpaces) {
      // output may lose some whitespaces, see https://youtrack.jetbrains.com/issue/IDEA-236807
      int trimmedSpaces = 80 - trimmed.length();
      if (trimmedSpaces > 0) {
        trimmed += GoStringUtil.repeat(" ", trimmedSpaces);
      }
    }
    myPendingInvalidJson.append(trimmed);
  }

  private static boolean looksLikeJson(@NotNull String text) {
    return text.startsWith("{\"Time\":");
  }

  @NotNull
  private static String decodeComparisonValue(@Nullable String value) {
    String literalText = GoStringUtil.unescapeStringLiteralText(value);
    return literalText != null ? literalText : GoStringUtil.notNullize(value);
  }

  private void processComparisonFailure(@Nullable String nodeId, @Nullable String output) {
    if (GoStringUtil.isNotEmpty(nodeId) && GoStringUtil.isNotEmpty(output)) {
      String trimmedOutput = GoStringUtil.trimLeading(output);
      ComparisonFailures comparisonFailures = myComparisonFailures.get(nodeId);
      if (comparisonFailures != null) {
        if (trimmedOutput.startsWith("expected:")) {
          comparisonFailures.expected = GoStringUtil.trimStart(trimmedOutput, "expected:").trim();
        }
        else if (trimmedOutput.startsWith("actual  :")) {
          comparisonFailures.actual = GoStringUtil.trimStart(trimmedOutput, "actual  :").trim();
        }
      }
      else if (Pattern.compile(NOT_EQUAL_PATTERN).matcher(trimmedOutput).matches()) {
        myComparisonFailures.put(nodeId, new ComparisonFailures());
      }
    }
  }

  private boolean startPackage(@NotNull GoTestEvent event, ServiceMessageVisitor visitor) {
    assert event.importPath != null;
    myEmptySuites.add(event.getNodeId());
    return startTest(event, event.importPath, event.importPath, null, event.getPackageLocationString(myDefaultImportPath), true, visitor);
  }

  private boolean startTest(@NotNull GoTestEvent event, ServiceMessageVisitor visitor) {
    NameAndParentTestName names = getNames(event.testName, event.importPath);
    assert names.testName != null;
    String parentNodeId = event.getParentNodeId(names.parentTestName);
    return startTest(event, names.testName, event.getNodeId(), parentNodeId, event.getTestLocationString(myDefaultImportPath), false, visitor);
  }

  private boolean output(@NotNull String testName,
                         @Nullable String nodeId,
                         @Nullable String output,
                         @NotNull OutputType outputType,
                         ServiceMessageVisitor visitor) {
    ServiceMessageBuilder message = outputType == OutputType.STDERR ? ServiceMessageBuilder.testStdErr(testName)
                                                                    : ServiceMessageBuilder.testStdOut(testName);
    message.attribute("nodeId", nodeId).attribute("out", output);
    return myProcessor.processEvent(message.build(), outputType, visitor);
  }

  private boolean startTest(@NotNull GoTestEvent event,
                            @NotNull String testName,
                            @Nullable String nodeId,
                            @Nullable String parentNodeId,
                            @NotNull String locationString,
                            boolean suite,
                            ServiceMessageVisitor visitor) {
    myStartedTests.add(nodeId);
    myEmptySuites.remove(parentNodeId);
    ServiceMessageBuilder startTest = suite ? ServiceMessageBuilder.testSuiteStarted(testName)
                                            : ServiceMessageBuilder.testStarted(testName);
    startTest
      .attribute("nodeId", nodeId)
      .attribute("parentNodeId", GoStringUtil.notNullize(parentNodeId, "0"))
      .attribute("locationHint", locationString)
      .attribute("message", GoStringUtil.notNullize(event.output));
    return myProcessor.processEvent(startTest.build(), OutputType.SYSTEM, visitor);
  }

  private void finishNonJsonOutputIfNeeded(ServiceMessageVisitor visitor) {
    synchronized (myFailTestSync) {
      if (myCurrentFailedImportPath != null) {
        String nodeId = myCurrentFailedImportPath;
        myCurrentFailedImportPath = null;
        ServiceMessageBuilder failTest = ServiceMessageBuilder.testFailed(nodeId)
          .attribute("nodeId", nodeId)
          .attribute("message", "");
        myProcessor.processEvent(failTest.build(), OutputType.SYSTEM, visitor);
        ServiceMessageBuilder finishTest = ServiceMessageBuilder.testSuiteFinished(nodeId).attribute("nodeId", nodeId);
        myProcessor.processEvent(finishTest.build(), OutputType.SYSTEM, visitor);
      }
    }
  }

  private boolean finishTest(@NotNull GoTestEvent event, @NotNull String testName, boolean suite, ServiceMessageVisitor visitor) {
    ServiceMessageBuilder finishTest = suite ? ServiceMessageBuilder.testSuiteFinished(GoStringUtil.notNullize(event.importPath))
                                             : ServiceMessageBuilder.testFinished(testName);
    finishTest.attribute("nodeId", event.getNodeId()).attribute("duration", event.getElapsedTime());
    return myProcessor.processEvent(finishTest.build(), OutputType.SYSTEM, visitor);
  }

  private boolean handleNonJsonOutput(@NotNull String text, @NotNull OutputType type, ServiceMessageVisitor visitor) {
    synchronized (myFailTestSync) {
      if (text.startsWith("# ")) {
        myCurrentFailedImportPath = GoStringUtil.trimStart(text, "# ");
        ServiceMessageBuilder messageBuilder = ServiceMessageBuilder.testSuiteStarted(myCurrentFailedImportPath)
          .attribute("nodeId", myCurrentFailedImportPath)
          .attribute("parentNodeId", "0");
        if (!myProcessor.processEvent(messageBuilder.build(), type, visitor)) {
          return false;
        }
      }
      if (myCurrentFailedImportPath != null) {
        return output(myCurrentFailedImportPath, myCurrentFailedImportPath, text, type, visitor);
      }
      return false;
    }
  }

  enum Action {RUN, PASS, SKIP, FAIL, OUTPUT, CONT, PAUSE}

  private NameAndParentTestName getNames(@Nullable String testName, @Nullable String importPath) {
    String parentTestName = "";
    if (testName != null) {
      String notNullImportPath = GoStringUtil.notNullize(importPath);
      int lastSlash = testName.lastIndexOf('/');
      while (lastSlash > -1) {
        String beforeSlash = testName.substring(0, lastSlash);
        if (lastSlash < testName.length() - 1 && myStartedTests.contains(notNullImportPath + "#" + beforeSlash)) {
          testName = testName.substring(lastSlash + 1);
          parentTestName = beforeSlash;
          break;
        }
        lastSlash = testName.lastIndexOf('/', lastSlash - 1);
      }
    }
    return new NameAndParentTestName(testName != null ? testName : importPath, parentTestName);
  }

  private static class ComparisonFailures {
    public String expected;
    public String actual;
  }

  private static final class NameAndParentTestName {
    public final String testName;
    public final String parentTestName;

    private NameAndParentTestName(@Nullable String testName, @Nullable String parentTestName) {
      this.testName = testName;
      this.parentTestName = parentTestName;
    }
  }

  @Nullable
  public Double getTotalTimeSeconds() {
    return myTotalTimeSeconds;
  }
}
